/* LZWCD
 * LZW compression and decompression utility
 * Copyright (c) 2009 Petr Koupy <petr.koupy@gmail.com>
 *
 * GNU General Public License Usage
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* -------------------------------------------------------------------------- */
% Main

% lzwc(+InputFile,+OutputFile).
lzwc(InputFile,OutputFile):-
  see(InputFile),
  tell(OutputFile),
  compress,
  seen,
  told.

% lzwd(+InputFile,+OutputFile).
lzwd(InputFile,OutputFile):-
  see(InputFile),
  tell(OutputFile),
  decompress,
  seen,
  told.
  
help:-
  nl,
  write('LZWCD'),
  nl,
  write('LZW compression and decompression utility'),
  nl,
  write('Copyright (c) 2009 Petr Koupy <petr.koupy@gmail.com>'),
  nl,
  nl,
  write('Usage:'),
  nl,
  write('Compression:   lzwc(+InputFile,+OutputFile).'),
  nl,
  write('Decompression: lzwd(+InputFile,+OutputFile).'),
  nl,
  nl.

/* -------------------------------------------------------------------------- */
% Input/Output

% getCode(-Code).
% Note that implementation of this predicate must correspond with maximum size
% of the dictionary (see initialize in Dictionary section).
% Current implementation works with 16 bit codes, thus every code is saved as
% lower order and upper order byte. Note that endianism of of this predicate
% must be the same as endianism of putCode/1 predicate.
getCode(Code):-
  peek_byte(Byte),
  Byte =:= -1,
  !,
  Code = Byte.
getCode(Code):-
  get_byte(Lower),
  get_byte(Upper),
  Code is Upper*256+Lower.

% putCode(+Code).
% Note that implementation of this predicate must correspond with maximum size
% of the dictionary (see initialize in Dictionary section).
% Current implementation works with 16 bit codes, thus every code is saved as
% lower order and upper order byte. Note that endianism of of this predicate
% must be the same as endianism of getCode/1 predicate.
putCode(Code):-
  Upper is Code//256,
  Lower is Code mod 256,
  put_byte(Lower),
  put_byte(Upper).
  
% putByteList(+ByteList).
putByteList([]):-
  !.
putByteList([FirstByte|Rest]):-
  put_byte(FirstByte),
  putByteList(Rest).

/* -------------------------------------------------------------------------- */
% Engine

/* Compressor pseudocode:
 *    w = NIL;
 *    add all possible charcodes to the dictionary
 *    for (every character c in the uncompressed data) do
 *        if ((w + c) exists in the dictionary) then
 *            w = w + c;
 *        else 
 *            add the dictionary code for w to output;
 *            add (w + c) to the dictionary;
 *            w = c;
 *        endif
 *    done
 *    add the dictionary code for w to output;
 *    display output;  
 */ 

compress:-
  initialize,
  get_byte(FirstByte),
  dictionary([FirstByte],FirstCode),
  get_byte(NextByte),
  compress(NextByte,[FirstByte],FirstCode),
  wipe.
  
% compress(+Byte,+LastByteList,+LastCode).
% All bytes are already encoded, just output the remaining code.
compress(CurrentByte,_,LastCode):-
  CurrentByte =:= -1,
  putCode(LastCode),
  !.
% Extend LastByteList with CurrentByte while there is some dictionary record 
% for this extension.
compress(CurrentByte,LastByteList,_):-
  append(LastByteList,[CurrentByte],CurrentByteList),
  dictionary(CurrentByteList,CurrentCode),
  !,
  get_byte(NextByte),
  compress(NextByte,CurrentByteList,CurrentCode).
% When there is no record for extended LastByteList, do three actions:
% 1) If there is still enough space in dictionary, add new dictionary record 
%    for extended LastByteList.
% 2) Output code for LastByteList.
% 3) Delete LastByteList and continue with CurrentByte from input.
compress(CurrentByte,LastByteList,LastCode):-
  append(LastByteList,[CurrentByte],CurrentByteList),
  not(dictionary(CurrentByteList,_)),
  next(NewCode),
  size(MaxCode),
  NewCode > MaxCode,
  !,
  putCode(LastCode),
  dictionary([CurrentByte],NewCurrentCode),
  get_byte(NextByte),
  compress(NextByte,[CurrentByte],NewCurrentCode).
compress(CurrentByte,LastByteList,LastCode):-
  append(LastByteList,[CurrentByte],CurrentByteList),
  not(dictionary(CurrentByteList,_)),
  next(NewCode),
  size(MaxCode),
  NewCode =< MaxCode,
  !,
  NewNextCode is NewCode+1,
  retract(next(NewCode)),
  asserta(next(NewNextCode)),
  assertz(dictionary(CurrentByteList,NewCode)),
  putCode(LastCode),
  dictionary([CurrentByte],NewCurrentCode),
  get_byte(NextByte),
  compress(NextByte,[CurrentByte],NewCurrentCode).
  
/* Decompressor pseudocode:
 *    add all possible charcodes to the dictionary
 *    read a char k;
 *    output k;
 *    w = k;
 *    while (read a char k) do
 *       if (index k exists in dictionary) then
 *           entry = dictionary entry for k;
 *       else if (k == currSizeDict)
 *           entry = w + w[0];
 *       else
 *           signal invalid code;
 *       endif
 *       output entry;
 *       add w+entry[0] to the dictionary;
 *       w = entry;
 *    done
 */ 
  
decompress:-
  initialize,
  getCode(FirstCode),
  dictionary([FirstByte],FirstCode),
  putByteList([FirstByte]),
  getCode(NextCode),
  decompress(NextCode,[FirstByte]),
  wipe.

% decompress(+Code,LastByteList).
% All bytes are already decoded.
decompress(CurrentCode,_):-
  CurrentCode =:= -1,
  !.
% Current code, which is not contained in dictionary, must be treated in a 
% special way. It follows from the nature of LZW that this happens if and only 
% if the next byte in the input is the same as the first byte of LastByteList.
% Thus the newly added dictionary record and output byte sequence are equal.
% Note that new record is added into dictionary only if there is enough free
% space in the dictionary.
decompress(CurrentCode,[LastFirst|LastRest]):-
  next(NewCode),
  NewCode =:= CurrentCode,
  size(MaxCode),
  NewCode > MaxCode,
  !,
  append([LastFirst|LastRest],[LastFirst],CurrentByteList),
  putByteList(CurrentByteList),
  getCode(NextCode),
  decompress(NextCode,CurrentByteList).
decompress(CurrentCode,[LastFirst|LastRest]):-
  next(NewCode),
  NewCode =:= CurrentCode,
  size(MaxCode),
  NewCode =< MaxCode,
  !,
  append([LastFirst|LastRest],[LastFirst],CurrentByteList),
  putByteList(CurrentByteList),
  NextNewCode is NewCode+1,
  retract(next(NewCode)),
  asserta(next(NextNewCode)),
  assertz(dictionary(CurrentByteList,NewCode)),
  getCode(NextCode),
  decompress(NextCode,CurrentByteList).
% Standard behaviour of decompress is to find and output CurrentCode in
% dictionary. Newly added dictionary record consists of LastByteList and first
% byte of currently recognized dictionary record. 
% Note that new record is added into dictionary only if there is enough free
% space in the dictionary.
decompress(CurrentCode,_):-
  next(NewCode),
  NewCode =\= CurrentCode,
  size(MaxCode),
  NewCode > MaxCode,
  !,
  dictionary(CurrentByteList,CurrentCode),
  putByteList(CurrentByteList),
  getCode(NextCode),
  decompress(NextCode,CurrentByteList).
decompress(CurrentCode,LastByteList):-
  next(NewCode),
  NewCode =\= CurrentCode,
  size(MaxCode),
  NewCode =< MaxCode,
  !,
  dictionary([CurrentFirst|CurrentRest],CurrentCode),
  putByteList([CurrentFirst|CurrentRest]),
  append(LastByteList,[CurrentFirst],NewCurrentByteList),
  NextNewCode is NewCode+1,
  retract(next(NewCode)),
  asserta(next(NextNewCode)),
  assertz(dictionary(NewCurrentByteList,NewCode)),
  getCode(NextCode),
  decompress(NextCode,[CurrentFirst|CurrentRest]).

/* -------------------------------------------------------------------------- */
% Dictionary

:-dynamic next/1.
:-dynamic size/1.
:-dynamic dictionary/2.

% Wipes dictionary.
wipe:-
  retractall(next(_)),
  retractall(size(_)),
  retractall(dictionary(_,_)).

% Builds initial dictionary of records dictionary(ByteList,Code). Term
% next(NewCode) contains lowest code which is not contained in current 
% dictionary. Term size(MaxCode) contains maximum code, which can be stored in
% the dictionary. Note that MaxCode must correspond with implementation of
% getCode/1 and putCode/1 in Input/Output section.
initialize:-
  wipe,
  asserta(next(256)),
  asserta(size(65535)),
  assertz(dictionary([0],0)),
  assertz(dictionary([1],1)),
  assertz(dictionary([2],2)),
  assertz(dictionary([3],3)),
  assertz(dictionary([4],4)),
  assertz(dictionary([5],5)),
  assertz(dictionary([6],6)),
  assertz(dictionary([7],7)),
  assertz(dictionary([8],8)),
  assertz(dictionary([9],9)),
  assertz(dictionary([10],10)),
  assertz(dictionary([11],11)),
  assertz(dictionary([12],12)),
  assertz(dictionary([13],13)),
  assertz(dictionary([14],14)),
  assertz(dictionary([15],15)),
  assertz(dictionary([16],16)),
  assertz(dictionary([17],17)),
  assertz(dictionary([18],18)),
  assertz(dictionary([19],19)),
  assertz(dictionary([20],20)),
  assertz(dictionary([21],21)),
  assertz(dictionary([22],22)),
  assertz(dictionary([23],23)),
  assertz(dictionary([24],24)),
  assertz(dictionary([25],25)),
  assertz(dictionary([26],26)),
  assertz(dictionary([27],27)),
  assertz(dictionary([28],28)),
  assertz(dictionary([29],29)),
  assertz(dictionary([30],30)),
  assertz(dictionary([31],31)),
  assertz(dictionary([32],32)),
  assertz(dictionary([33],33)),
  assertz(dictionary([34],34)),
  assertz(dictionary([35],35)),
  assertz(dictionary([36],36)),
  assertz(dictionary([37],37)),
  assertz(dictionary([38],38)),
  assertz(dictionary([39],39)),
  assertz(dictionary([40],40)),
  assertz(dictionary([41],41)),
  assertz(dictionary([42],42)),
  assertz(dictionary([43],43)),
  assertz(dictionary([44],44)),
  assertz(dictionary([45],45)),
  assertz(dictionary([46],46)),
  assertz(dictionary([47],47)),
  assertz(dictionary([48],48)),
  assertz(dictionary([49],49)),
  assertz(dictionary([50],50)),
  assertz(dictionary([51],51)),
  assertz(dictionary([52],52)),
  assertz(dictionary([53],53)),
  assertz(dictionary([54],54)),
  assertz(dictionary([55],55)),
  assertz(dictionary([56],56)),
  assertz(dictionary([57],57)),
  assertz(dictionary([58],58)),
  assertz(dictionary([59],59)),
  assertz(dictionary([60],60)),
  assertz(dictionary([61],61)),
  assertz(dictionary([62],62)),
  assertz(dictionary([63],63)),
  assertz(dictionary([64],64)),
  assertz(dictionary([65],65)),
  assertz(dictionary([66],66)),
  assertz(dictionary([67],67)),
  assertz(dictionary([68],68)),
  assertz(dictionary([69],69)),
  assertz(dictionary([70],70)),
  assertz(dictionary([71],71)),
  assertz(dictionary([72],72)),
  assertz(dictionary([73],73)),
  assertz(dictionary([74],74)),
  assertz(dictionary([75],75)),
  assertz(dictionary([76],76)),
  assertz(dictionary([77],77)),
  assertz(dictionary([78],78)),
  assertz(dictionary([79],79)),
  assertz(dictionary([80],80)),
  assertz(dictionary([81],81)),
  assertz(dictionary([82],82)),
  assertz(dictionary([83],83)),
  assertz(dictionary([84],84)),
  assertz(dictionary([85],85)),
  assertz(dictionary([86],86)),
  assertz(dictionary([87],87)),
  assertz(dictionary([88],88)),
  assertz(dictionary([89],89)),
  assertz(dictionary([90],90)),
  assertz(dictionary([91],91)),
  assertz(dictionary([92],92)),
  assertz(dictionary([93],93)),
  assertz(dictionary([94],94)),
  assertz(dictionary([95],95)),
  assertz(dictionary([96],96)),
  assertz(dictionary([97],97)),
  assertz(dictionary([98],98)),
  assertz(dictionary([99],99)),
  assertz(dictionary([100],100)),
  assertz(dictionary([101],101)),
  assertz(dictionary([102],102)),
  assertz(dictionary([103],103)),
  assertz(dictionary([104],104)),
  assertz(dictionary([105],105)),
  assertz(dictionary([106],106)),
  assertz(dictionary([107],107)),
  assertz(dictionary([108],108)),
  assertz(dictionary([109],109)),
  assertz(dictionary([110],110)),
  assertz(dictionary([111],111)),
  assertz(dictionary([112],112)),
  assertz(dictionary([113],113)),
  assertz(dictionary([114],114)),
  assertz(dictionary([115],115)),
  assertz(dictionary([116],116)),
  assertz(dictionary([117],117)),
  assertz(dictionary([118],118)),
  assertz(dictionary([119],119)),
  assertz(dictionary([120],120)),
  assertz(dictionary([121],121)),
  assertz(dictionary([122],122)),
  assertz(dictionary([123],123)),
  assertz(dictionary([124],124)),
  assertz(dictionary([125],125)),
  assertz(dictionary([126],126)),
  assertz(dictionary([127],127)),
  assertz(dictionary([128],128)),
  assertz(dictionary([129],129)),
  assertz(dictionary([130],130)),
  assertz(dictionary([131],131)),
  assertz(dictionary([132],132)),
  assertz(dictionary([133],133)),
  assertz(dictionary([134],134)),
  assertz(dictionary([135],135)),
  assertz(dictionary([136],136)),
  assertz(dictionary([137],137)),
  assertz(dictionary([138],138)),
  assertz(dictionary([139],139)),
  assertz(dictionary([140],140)),
  assertz(dictionary([141],141)),
  assertz(dictionary([142],142)),
  assertz(dictionary([143],143)),
  assertz(dictionary([144],144)),
  assertz(dictionary([145],145)),
  assertz(dictionary([146],146)),
  assertz(dictionary([147],147)),
  assertz(dictionary([148],148)),
  assertz(dictionary([149],149)),
  assertz(dictionary([150],150)),
  assertz(dictionary([151],151)),
  assertz(dictionary([152],152)),
  assertz(dictionary([153],153)),
  assertz(dictionary([154],154)),
  assertz(dictionary([155],155)),
  assertz(dictionary([156],156)),
  assertz(dictionary([157],157)),
  assertz(dictionary([158],158)),
  assertz(dictionary([159],159)),
  assertz(dictionary([160],160)),
  assertz(dictionary([161],161)),
  assertz(dictionary([162],162)),
  assertz(dictionary([163],163)),
  assertz(dictionary([164],164)),
  assertz(dictionary([165],165)),
  assertz(dictionary([166],166)),
  assertz(dictionary([167],167)),
  assertz(dictionary([168],168)),
  assertz(dictionary([169],169)),
  assertz(dictionary([170],170)),
  assertz(dictionary([171],171)),
  assertz(dictionary([172],172)),
  assertz(dictionary([173],173)),
  assertz(dictionary([174],174)),
  assertz(dictionary([175],175)),
  assertz(dictionary([176],176)),
  assertz(dictionary([177],177)),
  assertz(dictionary([178],178)),
  assertz(dictionary([179],179)),
  assertz(dictionary([180],180)),
  assertz(dictionary([181],181)),
  assertz(dictionary([182],182)),
  assertz(dictionary([183],183)),
  assertz(dictionary([184],184)),
  assertz(dictionary([185],185)),
  assertz(dictionary([186],186)),
  assertz(dictionary([187],187)),
  assertz(dictionary([188],188)),
  assertz(dictionary([189],189)),
  assertz(dictionary([190],190)),
  assertz(dictionary([191],191)),
  assertz(dictionary([192],192)),
  assertz(dictionary([193],193)),
  assertz(dictionary([194],194)),
  assertz(dictionary([195],195)),
  assertz(dictionary([196],196)),
  assertz(dictionary([197],197)),
  assertz(dictionary([198],198)),
  assertz(dictionary([199],199)),
  assertz(dictionary([200],200)),
  assertz(dictionary([201],201)),
  assertz(dictionary([202],202)),
  assertz(dictionary([203],203)),
  assertz(dictionary([204],204)),
  assertz(dictionary([205],205)),
  assertz(dictionary([206],206)),
  assertz(dictionary([207],207)),
  assertz(dictionary([208],208)),
  assertz(dictionary([209],209)),
  assertz(dictionary([210],210)),
  assertz(dictionary([211],211)),
  assertz(dictionary([212],212)),
  assertz(dictionary([213],213)),
  assertz(dictionary([214],214)),
  assertz(dictionary([215],215)),
  assertz(dictionary([216],216)),
  assertz(dictionary([217],217)),
  assertz(dictionary([218],218)),
  assertz(dictionary([219],219)),
  assertz(dictionary([220],220)),
  assertz(dictionary([221],221)),
  assertz(dictionary([222],222)),
  assertz(dictionary([223],223)),
  assertz(dictionary([224],224)),
  assertz(dictionary([225],225)),
  assertz(dictionary([226],226)),
  assertz(dictionary([227],227)),
  assertz(dictionary([228],228)),
  assertz(dictionary([229],229)),
  assertz(dictionary([230],230)),
  assertz(dictionary([231],231)),
  assertz(dictionary([232],232)),
  assertz(dictionary([233],233)),
  assertz(dictionary([234],234)),
  assertz(dictionary([235],235)),
  assertz(dictionary([236],236)),
  assertz(dictionary([237],237)),
  assertz(dictionary([238],238)),
  assertz(dictionary([239],239)),
  assertz(dictionary([240],240)),
  assertz(dictionary([241],241)),
  assertz(dictionary([242],242)),
  assertz(dictionary([243],243)),
  assertz(dictionary([244],244)),
  assertz(dictionary([245],245)),
  assertz(dictionary([246],246)),
  assertz(dictionary([247],247)),
  assertz(dictionary([248],248)),
  assertz(dictionary([249],249)),
  assertz(dictionary([250],250)),
  assertz(dictionary([251],251)),
  assertz(dictionary([252],252)),
  assertz(dictionary([253],253)),
  assertz(dictionary([254],254)),
  assertz(dictionary([255],255)).
